/*
   here's the short plan of how this will be done:

   Read the command line, get the supplied dump level as well as Volume Set.

   Find the Latest Dump that we can try to use incrementals off of.

   Expand the volumeset into its component list of volumes.

   Find the date the volume was last dumped. This is the time from
   which the incremental should be done, if any.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <afs/param.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <sys/stat.h>
#include <rx/xdr.h>
#include <rx/rx.h>
#include <afs/afsint.h>
#include <afs/nfs.h>
#include <afs/errors.h>
#include <lock.h>
#include <lwp.h>
#include <afs/ihandle.h>
#include <afs/dir.h>
#include <afs/vnode.h>
#include <afs/volume.h>
#include <afs/partition.h>
#include <afs/fssync.h>
#include <afs/acl.h>
#include <afs/vldbint.h>
#include <afs/dirpath.h>
#include <afs/volser.h>
#include <afs/volint.h>
#include <afs/assert.h>
#include <afs/cellconfig.h>

#include "config.h"

#define EIGHT_BYTE_INT long long

void DFlushVolume(void) {} ;

EIGHT_BYTE_INT afs_estimate(char *volname, time_t prev_dump);

EIGHT_BYTE_INT SumVnodeIndex(Volume *vp, VnodeClass class, afs_int32 prev_dump);

int main(int argc, char *argv[])
{
   int code = 0;
   char *volname;
   int dlevel;
   EIGHT_BYTE_INT size;
   char prev_dumpstr[80];
   time_t prev_dump;
   
   /*
     Read the command line, get the supplied dump level as well as a Volume.
   */
   if(argc<3)
   {
      fprintf(stderr, "%s\nRun this utility as follows:\n\t%s date volume_name\n", AMANDA_AFS_VERSION, argv[0]);
      fprintf(stderr, "date specified in seconds since EPOCH.\n");
      fprintf(stderr, "volume_name -- self-explanatory.\n");
      return -1;
   }
   
   /* get the dump level */
   sscanf(argv[1], "%ld", &prev_dump);
   /* get the volume name */
   volname = strdup(argv[2]);

   /*
     now go through the volume structures and fish out the needed
     things. 
   */
   
   size = afs_estimate(volname, prev_dump);

   printf("%lld\n", size);
   
   free(volname);
   return code;
}

/*
 * EIGHT_BYTE_INT afs_estimate()
 * 
 * Parameters: volname   - name of the volume to estimate
 *             prev_dump - the time from which to include vnodes.
 *
 * Function: Examines the volume 'volname' and sums up the sizes of
 *           all vnodes, whose modification date is after
 *           'prev_dump'. 
 * 
 * 06/06/01 10:53:35
 * 
 */
EIGHT_BYTE_INT afs_estimate(char *volname, time_t prev_dump)
{
   struct nvldbentry entry;
   extern struct ubik_client *cstruct;
   Volume *vp;
   afs_int32 id;
   afs_int32 partition;
   afs_int32 code;
   afs_int32 sauth;
   Error err;
   char *tcell;
   int i, idx, length;
   EIGHT_BYTE_INT dump_size = 0;
   int rxpackets = 100;
   static char nullString[1] = "";
   extern int UV_SetSecurity();

   VInitVolumePackage(volumeUtility, 0, 0, CONNECT_FS, 0); 
/*   DInit(40); */
   
   
   /* Initialize the ubik_client connection */
   rx_SetRxDeadTime(90);
   cstruct = (struct ubik_client *)0;

   sauth = 0;
   tcell = (char *) 0;
   if (code = vsu_ClientInit(1, AFSDIR_CLIENT_ETC_DIRPATH, tcell,
			     sauth, &cstruct, UV_SetSecurity)) {
      fprintf(STDERR,"could not initialize VLDB library (code=%u) \n",code);
      exit(1);
   }
  
   bzero(&entry, sizeof(struct nvldbentry));
   /* look up the VolId and partition */
   code = VLDB_GetEntryByName(volname, &entry);
   if(!code) {
      /* find out which volume we've been asked to deal with */
      
      length = strlen(volname);
      if(!strcmp(&volname[length - 9],".readonly")) {
	 idx = ROVOL;
      }
      else if(!strcmp(&volname[length - 7 ],".backup")) {
	 idx = BACKVOL;
      }
      else {
	 idx = RWVOL;
      }
      
      id = entry.volumeId[idx]; /* THIS IS VERY IMPORTANT!!!! */

      /* if 'vos dump volname.backup' is used, then above should say
         BACKVOL, otherwise, should say RWVOL. */
      
   } else {
      printf ("D'oh!\n");
   }

   vp = VAttachVolume(&err, id, V_DUMP);
   if (err) {
      /* give up */
      if (vp) VDetachVolume(&err, vp);
   }
   else
   {
      /* add size of the DumpDumpHeader to dump_size */
      
      dump_size = 9 + 5 + (2+strlen(V_name(vp))) + (3+4*2);
      
      /* do the equivalent of DumpPartial and add its result to dump_size */      
      
      dump_size += 1 + 5 + 5 + (2+strlen(V_name(vp))) + 2 + 2 + 5 + 2 + 5 + 5 + 5 + 5 + 5 +
	 5 + 5 + 5 + 5 + 5 + 5 + 5 + 5 + (2+strlen(V_offlineMessage(vp))) +
	 (2+strlen(nullString)) + (3+4*(sizeof(V_weekUse(vp))/sizeof(V_weekUse(vp)[0]))) + 5 + 5;
      
      /* add the size of DumpEnd to dump_size */
      dump_size += 5;
      
      dump_size += SumVnodeIndex(vp, vSmall, (afs_int32)prev_dump);
      dump_size += SumVnodeIndex(vp, vLarge, (afs_int32)prev_dump);
    
   }
   
   if (vp) VDetachVolume(&err, vp);
   
   if (cstruct) {
      ubik_ClientDestroy (cstruct);
      cstruct = 0;
   }
   rx_Finalize();
   
   return (EIGHT_BYTE_INT)dump_size;
}

/*
 * EIGHT_BYTE_INT SumVnodeIndex()
 * 
 * Parameters: vp        - the volume which to estimate
 *             class     - the class of vnodes in 'vp' to examine
 *             prev_dump - the time of the last dump.
 *
 * Function: This is almost verbatim DumpVnodeIndex. Rummage through
 *           the vnodes of class 'class' in 'vp' and sum their
 *           lengths. 
 * 
 * 06/06/01 11:02:43
 * 
 */
EIGHT_BYTE_INT SumVnodeIndex(Volume *vp, VnodeClass class, afs_int32 prev_dump)
{
   register struct VnodeClassInfo *vcp = &VnodeClassInfo[class];
   char buf[SIZEOF_LARGEDISKVNODE];
   struct VnodeDiskObject *vnode = (struct VnodeDiskObject *) buf;
   StreamHandle_t *file, *file1;
   IHandle_t *handle,*ihP1;
   FdHandle_t *fdP, *fdP1;
   int size, flag;
   unsigned char *filedata;
   register int vnodeIndex, nVnodes;
   struct DirHeader *dhp;
   struct DirEntry *ep;
   int offset, hashbucket;
   EIGHT_BYTE_INT vnode_est = 0;
   
   fdP = IH_OPEN(vp->vnodeIndex[class].handle);
   assert(fdP != NULL);
   file = FDH_FDOPEN(fdP, "r+");
   assert(file != NULL);
   size = OS_SIZE(fdP->fd_fd);
   assert(size != -1);
   nVnodes = (size / vcp->diskSize) - 1;
   if (nVnodes > 0) {
      assert((nVnodes+1)*vcp->diskSize == size)
      assert(STREAM_SEEK(file, vcp->diskSize, 0) == 0)
   }
   else nVnodes = 0;
   for (vnodeIndex = 0;
	nVnodes && STREAM_READ(vnode, vcp->diskSize, 1, file) == 1;
	nVnodes--, vnodeIndex++)
   {
      
      if(vnode->type == vNull) continue; 
      
      if ((class == vLarge) || (vnode->serverModifyTime >= prev_dump))
      {
	 vnode_est += 8 + 2 + 3 + 5 + 5 + 5 +  5 + 5 + 3 + 5 + 5 + 1;
	 
	 if(vnode->group)
	 {
	    vnode_est += 5;
	 }
	 
	 if(vnode->type == vDirectory)
	 {
	    vnode_est += (1 + VAclDiskSize(vnode));
	 }
	 
	 vnode_est += vnode->length;
      }
      else
      {
	 vnode_est += 9;
      }

   }
   STREAM_CLOSE(file);
   FDH_CLOSE(fdP);
   
   return vnode_est;
}

/*
 * int osi_audit())
 * 
 * Parameters: 
 * 
 * Function: Copied _verbatim_ out of vos.c.
 *           Makes compiles go happy.
 * 
 * 06/06/01 14:18:02
 * 
 */
int osi_audit()
{
/* this sucks but it works for now.
*/
return 0;
}
