#include "volset.h"
#include "volset-link.h"
#include "volset-helper.h"
#include "volset-dump.h"

int DumpVolSetHeader(int out, Node **vols)
{
   /* dump the volumeset header */
   
   return 0;
}

EIGHT_BYTE_INT DumpVolume(int out, struct vol *cur_vol, time_t prevdump, ONE_BYTE_INT last_vol)
{
   /* dump the volume in chunks */
   FILE *vos_pipe = NULL; /* pipe to read from vos dump */
   char cmd[1024];

   int bytes_read = 0;
   int code;
   struct tm *gmtm;
   int i;

   EIGHT_BYTE_INT data_written = 0;
   
   char prevtime[32];

   EIGHT_BYTE_INT chunk_number = 0;
   ONE_BYTE_INT last_chunk=0;
   ONE_BYTE_INT checksum;
   
   EIGHT_BYTE_INT t8;
   FOUR_BYTE_INT t4;
   FOUR_BYTE_INT s4;
   TWO_BYTE_INT t2;
   ONE_BYTE_INT t1;
   ONE_BYTE_INT *buf; /* buffer for the I/O */
   ONE_BYTE_INT *header; /* header for the chunks */
   
   gmtm = localtime(&prevdump);
   sprintf(prevtime, "%02d/%02d/%04d %2d:%02d:%02d",
	   gmtm->tm_mon+1, gmtm->tm_mday, gmtm->tm_year + 1900,
	   gmtm->tm_hour, gmtm->tm_min, gmtm->tm_sec);
   
/*   cmd = (char *) malloc(strlen(VOS_CMD) + strlen(" dump -time \"\" -localauth 2> /dev/null") + \
			 strlen(prevtime) + strlen(cur_vol->volname) + 1);
*/
   
   sprintf(cmd, "%s dump %s -time \"%s\" -localauth 2> /dev/null", VOS_CMD, cur_vol->volname, prevtime);

   vos_pipe = popen(cmd, "r");
   if(!vos_pipe)
   {
      fprintf(stderr, "Error: in popen: %s\n", strerrno(errno));
   }
/*
  if(cmd)
   {
      free(cmd);
      cmd = NULL;
   }
*/
   /*
     This loop handles the chunking of things into CHUNK_SIZE'd packages,
     i.e. one iteration of the loop == one chunk.
   */
   buf = (ONE_BYTE_INT *) malloc(CHUNK_SIZE*sizeof(ONE_BYTE_INT));
   header = (ONE_BYTE_INT *) malloc((CHUNK_HEADER_SIZE+(strlen(cur_vol->volname)+1))*sizeof(ONE_BYTE_INT));
   while(vos_pipe && !feof(vos_pipe) && !ferror(vos_pipe))
   {
      memset(buf, 0, CHUNK_SIZE);
      memset(header, 0, CHUNK_HEADER_SIZE+(strlen(cur_vol->volname)+1));

      bytes_read = fread(buf, 1, CHUNK_SIZE, vos_pipe); 

      if(bytes_read<CHUNK_SIZE && feof(vos_pipe))
      {
	 /* end of dump, pad it and ship it off! */
	 last_chunk = 1;
      }
      else if(bytes_read<CHUNK_SIZE && ferror(vos_pipe))
      {
	 /* got an I/O error, do something strange! */
	 fprintf(stderr, "Error: got an I/O error on read: %d\n", errno);
	 break;
      }
      
      
      /* Everything in network byte order, except single bytes (sic!) */
      t4 = htonl(CHUNK_MAGIC);
      memcpy(header+MAGIC_OFFSET, &t4, MAGIC_SIZE); /* magic */

      t1 = VERSION_MAJOR;
      memcpy(header+VMAJ_OFFSET, &t1, VMAJ_SIZE);  /* version major */
      
      t1 = VERSION_MINOR;
      memcpy(header+VMIN_OFFSET, &t1, VMIN_SIZE);  /* version minor */
      
      t2 = htons(CHUNK_HEADER_SIZE+(strlen(cur_vol->volname)+1));
      memcpy(header+HS_OFFSET, &t2, HS_SIZE); /* header size */

      t1 = last_vol;
      memcpy(header+LV_OFFSET, &t1, LV_SIZE);  /* last volume? */

      t1 = last_chunk;
      memcpy(header+LC_OFFSET, &t1, LC_SIZE); /* last chunk? */
      
      t8 = htonl(chunk_number);
      memcpy(header+CN_OFFSET, &t8, CN_SIZE);  /* chunk number */
      
      t4 = htonl(bytes_read);
      memcpy(header+DS_OFFSET, &t4, DS_SIZE);  /* data size */

/* put the support for partition and the server here */
      t4 = htonl(cur_vol->server);
      memcpy(header+SERV_OFFSET, &t4, SERV_SIZE);  /* server */
      
      t4 = htonl(cur_vol->partition);
      memcpy(header+PART_OFFSET, &t4, PART_SIZE);  /* partiton */ 
     
      t2 = htons(strlen(cur_vol->volname)+1);
      memcpy(header+VNS_OFFSET, &t2, VNS_SIZE);  /* volume name length */
      
      memcpy(header+VN_OFFSET, cur_vol->volname, strlen(cur_vol->volname)); /* volume name */
      memcpy(header+VN_OFFSET+strlen(cur_vol->volname)+1, "\000", 1);

      checksum = 0;
      
      t1 = CheckSum(header, CHUNK_HEADER_SIZE+(strlen(cur_vol->volname)+1));
      memcpy(header+VN_OFFSET+(strlen(cur_vol->volname)+1), &t1, HC_SIZE); /* checksum */
      
      /* write the actual header to file! */
      code = write(out, header, CHUNK_HEADER_SIZE+(strlen(cur_vol->volname)+1));
      if(code != CHUNK_HEADER_SIZE+(strlen(cur_vol->volname)+1))
      {
	 fprintf(stderr, "Error: could not write header to output!!!\n");
	 break;
      }

      data_written += code;
      
      code = write(out,buf, bytes_read);
      if(code != bytes_read)
      {
	 fprintf(stderr, "Error: could not write data to output!!!\n");
	 break;
      }

      data_written += code;

      /* now take care of the chunk_number */
      chunk_number++;
   }

   if(vos_pipe)
   {
      code = pclose(vos_pipe);
      if(code == -1)
      {
	 fprintf(stderr, "Error: error in closing the pipe!\n");
      }
   }
   if(buf)
   {
      free(buf);
      buf = NULL;
   }
   if(header)
   {
      free(header);
      header = NULL;
   }
   return data_written;
}
