#include "volset.h"
#include "volset-helper.h"
#include "volset-estimate.h"

/*
   here's the short plan of how this will be done:

   Read the command line, get the supplied dump level as well as Volume Set.

   Find the Latest Dump that we can try to use incrementals off of.

   Expand the volumeset into its component list of volumes.

   Find the date the volume was last dumped. This is the time from
   which the incremental should be done, if any.

*/


/*
 * void last_backup()
 * 
 * Parameters:
 *              vol_name     - determines which volume to look up
 *              level        - which dump level to look up
 *              prev_dumpstr - the date string that is returned
 *              prev_dump    - the actual date that is returned.
 *
 * Function: last_backup looks up the last time a backup of level
 *           'level' has been done on volume 'vol_name' and returns
 *           the date string in 'prev_dumpstr' and the time in
 *           'prev_dump'. This is borrowed from
 *           client-src/sendbackup-gnutar.c.
 * 
 * 06/05/01 17:39:00
 *
 */

/*
void last_backup(char *vol_name, int level, char *prev_dumpstr,
		 time_t *prev_dump)
{
   amandates_t *amdates;
   struct tm *gmtm;
   int l;

   
   if(!start_amandates(0))
      error("error [opening %s: %s]", AMANDATES_FILE, strerror(errno));

   amdates = amandates_lookup(vol_name);
   
   *prev_dump = (time_t)0;
   for(l = 0; l < level; l++) {
      if(amdates->dates[l] > *prev_dump)
	 *prev_dump = amdates->dates[l];
   }
   finish_amandates();
   free_amandates();

   gmtm = localtime(prev_dump);

   sprintf(prev_dumpstr, "%04d-%02d-%02d %2d:%02d:%02d GMT",
	    gmtm->tm_year + 1900, gmtm->tm_mon+1, gmtm->tm_mday,
	    gmtm->tm_hour, gmtm->tm_min, gmtm->tm_sec);

   sprintf(prev_dumpstr, "%02d/%02d/%04d %2d:%02d:%02d",
	   gmtm->tm_mon+1, gmtm->tm_mday, gmtm->tm_year + 1900,
	   gmtm->tm_hour, gmtm->tm_min, gmtm->tm_sec);
}
*/

/*
 * EIGHT_BYTE_INT EstimateVolume()
 * 
 * Parameters: cur_vol   - the volume to estimate
 *             prev_dump - the time from which to include vnodes.
 *
 * Function: Examines the volume 'cur_vol' and sums up the sizes of
 *           all vnodes, whose modification date is after
 *           'prev_dump'. 
 * 
 * 06/06/01 10:53:35
 * 
 */
EIGHT_BYTE_INT EstimateVolume(struct vol *cur_vol, time_t prev_dump)
{
   struct nvldbentry entry;
   Volume *vp;
   afs_int32 id;
   afs_int32 partition;
   afs_int32 code;
   afs_int32 sauth;
   Error err;
   int i, idx, length;
   EIGHT_BYTE_INT dump_size = 0;
   EIGHT_BYTE_INT t = 0;
   static char nullString[1] = "";
   
   bzero(&entry, sizeof(struct nvldbentry));
   /* look up the VolId and partition */
   /* find out which volume we've been asked to deal with */
      
   length = strlen(cur_vol->volname);
   if(!strcmp(&(cur_vol->volname)[length - 9],".readonly")) {
      idx = ROVOL;
   }
   else if(!strcmp(&(cur_vol->volname)[length - 7 ],".backup")) {
      idx = BACKVOL;
   }
   else {
      idx = RWVOL;
   }
   
   id = cur_vol->volumeId[idx]; /* THIS IS VERY IMPORTANT!!!! */
   
   /* if 'vos dump volname.backup' is used, then above should say
      BACKVOL, otherwise, should say RWVOL. */
   
   
   vp = VAttachVolume(&err, id, V_DUMP);
   if (err) {
      /* give up */
      if (vp) VDetachVolume(&err, vp);
   }
   else
   {
      /* add size of the DumpDumpHeader to dump_size */
      
      dump_size = 9 + 5 + (2+strlen(V_name(vp))) + (3+4*2);
      
      /* do the equivalent of DumpPartial and add its result to dump_size */
      
      dump_size += 1 + 5 + 5 + (2+strlen(V_name(vp))) + 2 + 2 + 5 + 2 + 5 + 5 + 5 + 5 + 5 +
	 5 + 5 + 5 + 5 + 5 + 5 + 5 + 5 + (2+strlen(V_offlineMessage(vp))) +
	 (2+strlen(nullString)) + (3+4*(sizeof(V_weekUse(vp))/sizeof(V_weekUse(vp)[0]))) + 5 + 5;
      
      dump_size += SumVnodeIndex(vp, vSmall, (afs_int32)prev_dump);
      dump_size += SumVnodeIndex(vp, vLarge, (afs_int32)prev_dump);
      
      /* add the size of DumpEnd to dump_size */
      dump_size += 5;
   }
   
   if (vp) VDetachVolume(&err, vp);

   /* now adjust for the format of the dump */

   if(dump_size < CHUNK_SIZE)
   {
      dump_size += CHUNK_HEADER_SIZE + length + 1;
   }
   else
   {
      /* do the chunking */
      t = dump_size / CHUNK_SIZE;
      dump_size += (CHUNK_HEADER_SIZE + length + 1)*(t+1);
   }
   
   return (EIGHT_BYTE_INT)dump_size;
}

/*
 * EIGHT_BYTE_INT SumVnodeIndex()
 * 
 * Parameters: vp        - the volume which to estimate
 *             class     - the class of vnodes in 'vp' to examine
 *             prev_dump - the time of the last dump.
 *
 * Function: This is almost verbatim DumpVnodeIndex. Rummage through
 *           the vnodes of class 'class' in 'vp' and sum their
 *           lengths. 
 * 
 * 06/06/01 11:02:43
 * 
 */
EIGHT_BYTE_INT SumVnodeIndex(Volume *vp, VnodeClass class, afs_int32 prev_dump)
{
   register struct VnodeClassInfo *vcp = &VnodeClassInfo[class];
   char buf[SIZEOF_LARGEDISKVNODE];
   struct VnodeDiskObject *vnode = (struct VnodeDiskObject *) buf;
   StreamHandle_t *file, *file1;
   IHandle_t *handle,*ihP1;
   FdHandle_t *fdP, *fdP1;
   int size, flag;
   unsigned char *filedata;
   register int vnodeIndex, nVnodes;
   struct DirHeader *dhp;
   struct DirEntry *ep;
   int offset, hashbucket;
   EIGHT_BYTE_INT vnode_est = 0;
   
   fdP = IH_OPEN(vp->vnodeIndex[class].handle);
   assert(fdP != NULL);
   file = FDH_FDOPEN(fdP, "r+");
   assert(file != NULL);
   size = OS_SIZE(fdP->fd_fd);
   assert(size != -1);
   nVnodes = (size / vcp->diskSize) - 1;
   if (nVnodes > 0) {
      assert((nVnodes+1)*vcp->diskSize == size)
      assert(STREAM_SEEK(file, vcp->diskSize, 0) == 0)
   }
   else nVnodes = 0;
   for (vnodeIndex = 0;
	nVnodes && STREAM_READ(vnode, vcp->diskSize, 1, file) == 1;
	nVnodes--, vnodeIndex++)
   {
      
      if(vnode->type == vNull) continue; 
      
      if ((class == vLarge) || (vnode->serverModifyTime >= prev_dump))
      {
	 vnode_est += 8 + 2 + 3 + 5 + 5 + 5 +  5 + 5 + 3 + 5 + 5 + 1;
	 
	 if(vnode->group)
	 {
	    vnode_est += 5;
	 }
	 
	 if(vnode->type == vDirectory)
	 {
	    vnode_est += (1 + VAclDiskSize(vnode));
	 }
	 
	 vnode_est += vnode->length;
      }
      else
      {
	 vnode_est += 9;
      }
   }
   STREAM_CLOSE(file);
   FDH_CLOSE(fdP);
   
   return vnode_est;
}
