#include "volset.h"
#include "volset-link.h"
#include "volset-helper.h"
#include "volset-dump.h"
#include "volset-restore.h"

int ReadVolSetHeader(FILE *in, Node **vols)
{
   /* dump the volumeset header */
   
   return 0;
}

int RestoreVolume(FILE *in, char *fname, char *volname)
{
   FILE *out;
/*   int out; */
   EIGHT_BYTE_INT t8;
   FOUR_BYTE_INT t4;
   TWO_BYTE_INT t2;
   ONE_BYTE_INT t1;

   EIGHT_BYTE_INT chunk_number, ch_num;
   FOUR_BYTE_INT dsize;
   TWO_BYTE_INT vnsize, hsize;
   ONE_BYTE_INT vmaj, checksum, csum;
   ONE_BYTE_INT last_vol, last_chunk;

   ONE_BYTE_INT *buf;  /* buffer for chunk data */
   ONE_BYTE_INT *header; /* header for the chunks */
   char *vname; /* volume name */
   afs_int32 orig_server, orig_partition;
   char *fileout;
   char *server, *partition;
   int i, done, bytes_read, code;
   int afs = 0;
   int error = 0;
      
   /* read chunk and if it is ok, spit it out into the appropriate file */
   
   done = 0;
   chunk_number = 0;
/*   out = -1; */
   out = NULL;
   buf = (ONE_BYTE_INT *) malloc(BUF_SIZE*sizeof(ONE_BYTE_INT));
   while(!done)
   {
      /* read header */
      
      bytes_read = 0;
      hsize = 0;

      t4 = htonl(CHUNK_MAGIC);
      if(!fsearch(in, (ONE_BYTE_INT *)&t4, MAGIC_SIZE)) /* magic */
      {
	 fprintf(stderr, "Error: Magic is bad!!!\n");
	 break;
      }
      bytes_read += MAGIC_SIZE;

      code = fread(&t1, 1, VMAJ_SIZE, in);  /* version minor */
      if(code != VMAJ_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      else if(t1 != VERSION_MAJOR)
      {
	 fprintf(stderr, "Error: wrong minor version!!!\n");
	 break;
      }
      vmaj = t1;
      bytes_read += code;

      code = fread(&t1, 1, VMIN_SIZE, in);  /* version minor */
      if(code != VMIN_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      else if(t1 != VERSION_MINOR)
      {
	 fprintf(stderr, "Error: wrong minor version!!!\n");
	 break;
      }
      bytes_read += code;
      
      code = fread(&t2, 1, HS_SIZE, in); /* header size */ 
      if(code != HS_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      hsize = ntohs(t2);
      bytes_read += code;

      /* fill the header */
      header = (ONE_BYTE_INT *) malloc(hsize*sizeof(ONE_BYTE_INT));
      memset(header, 0, hsize);
      memcpy(header+MAGIC_OFFSET, &t4, MAGIC_SIZE); /* magic */
      memcpy(header+HS_OFFSET, &t2, HS_SIZE); /* header size */
      memcpy(header+VMAJ_OFFSET, &vmaj, VMAJ_SIZE);  /* version major */
      memcpy(header+VMIN_OFFSET, &t1, VMIN_SIZE);  /* version minor */
         
      code = fread(&t1, 1, LV_SIZE, in);  /* last volume? */
      if(code != LV_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      memcpy(header+LV_OFFSET, &t1, LV_SIZE);  /* last volume? */
      last_vol = t1;
      bytes_read += code;

      code = fread(&t1, 1, LC_SIZE, in); /* last chunk? */
      if(code != LC_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      memcpy(header+LC_OFFSET, &t1, LC_SIZE); /* last chunk? */
      last_chunk = t1;
      bytes_read += code;
      
      code = fread(&t8, 1, CN_SIZE, in);  /* chunk number */
      if(code != CN_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      ch_num = ntohl(t8);
      memcpy(header+CN_OFFSET, &t8, CN_SIZE);  /* chunk number */
      bytes_read += code;

      code = fread(&t4, 1, DS_SIZE, in);  /* data size */
      if(code != DS_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      memcpy(header+DS_OFFSET, &t4, DS_SIZE);  /* data size */
      dsize = ntohl(t4);
      bytes_read += code;

/* put the support for partition and server here */ 
      code = fread(&t4, 1, SERV_SIZE, in);  /* server */
      if(code != SERV_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      memcpy(header+SERV_OFFSET, &t4, SERV_SIZE);  /* server */
      orig_server = ntohl(t4); /* to be checked against server.sin_addr.s_addr */
      bytes_read += code;
      
      code = fread(&t4, 1, PART_SIZE, in);  /* parition */
      if(code != PART_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      memcpy(header+PART_OFFSET, &t4, PART_SIZE);  /* parition */
      orig_partition = ntohl(t4); /* to be checked against partition */
      bytes_read += code;

      code = fread(&t2, 1, VNS_SIZE, in);  /* volume name length */
      if(code != VNS_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      memcpy(header+VNS_OFFSET, &t2, VNS_SIZE);  /* volume name length */
      vnsize = ntohs(t2);
      bytes_read += code;

      vname = (ONE_BYTE_INT *)malloc(vnsize*sizeof(ONE_BYTE_INT));
      
      code = fread(vname, 1, vnsize, in); /* volume name */
      if(code != vnsize)
      {
	 fprintf(stderr, "Error: could not read from file!!!\n");
	 break;
      }
      memcpy(header+VN_OFFSET, vname, vnsize); /* volume name */
      bytes_read += code;
      
      code = fread(&t1, 1, HC_SIZE, in); /* checksum */
      if(code != HC_SIZE)
      {
	 fprintf(stderr, "Error: could not read from file: %d\n", errno);
	 break;
      }
      bytes_read += code;
      csum = t1;
      checksum = CheckSum(header, CHUNK_HEADER_SIZE+vnsize);
      memcpy(header+VN_OFFSET+vnsize, &checksum, HC_SIZE); /* checksum */

      if(strcmp(vname, volname) == 0 || strcmp(volname, "*") == 0)
      {
	 if(chunk_number != ch_num)
	 {
	    fprintf(stderr, "Error: volume %s, chunk %ld, expected %ld -- chunks out of order!!!\n", vname, ch_num, chunk_number);
	    /* skip the data */
/*	    if(lseek(in, dsize, SEEK_CUR) == -1) */
	    if(fileseek(in, dsize, SEEK_CUR) == -1)
	    {
	       fprintf(stderr, "Error: could not seek the input file!\n");
	       break;
	    }
	 }
	 else if((hsize == bytes_read) && (checksum == csum))
	 {
	    chunk_number++;

/*	    if(out == -1) */
	    if(out == NULL)
	    {
	       fprintf(stderr, "Restoring: %s\n", vname);
	    }
	    
	    if(strcmp(fname, "-") == 0)
	    {
/*	       out = fileno(stdout);*/
	       out = stdout;
	    }
	    else if(out == NULL && strncmp(fname, "afs:", 4) == 0)
	    {
	       int k;
	       /* been asked to restore to a specific parition */

	       /* build the call to 'vos restore', call popen on it */
	       server = (char *) malloc(strlen(fname));
	       memset(server, 0, strlen(fname));
	       partition = (char *) malloc(strlen(fname));
	       memset(partition, 0, strlen(fname));
	       for(k = 3; k<strlen(fname) && fname[k] != '/'; k++);
	       memcpy(server, fname+4, k-4);
	       memcpy(partition, fname+k+1, strlen(fname)-k-1);
	       fileout = (char *) malloc(strlen(VOS_CMD) + strlen(" restore -name .r -partition  -server  -localauth") +
					 strlen(vname) + strlen(partition) + strlen(server) + 1);
	       sprintf(fileout, "%s restore -name %s.r -partition %s -server %s -localauth\000", VOS_CMD,
		       vname, partition, server);
	       out = popen(fileout, "w");
	       afs = 1;
	       free(partition);
	       free(server);
	       free(fileout);
	    }
/*	    else if(out == -1) */
	    else if(out == NULL)
	    {
	       /* open the new file */
	       if(strcmp(volname, "*") == 0)
	       {
		  fileout = (char *) malloc(strlen(fname)+1+strlen(vname));
		  sprintf(fileout, "%s/%s", fname, vname);
	       }
	       else
	       {
		  fileout = strdup(fname);
	       }
	       
/*	       out = open(fileout, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR); */
	       out = fopen(fileout, "w");
	       afs = 0;
/*	       if(out == -1) */
	       if(out == NULL)
	       {
/*		  out = open(fileout, O_WRONLY | O_TRUNC); */
/*		  if(out == -1)
		  { */
		     fprintf(stderr, "Error: could not open output file: %d\n", errno);
		     free(fileout);
		     break;
/*		  }
		  free(fileout); */
	       }
	    }
	    while(dsize > BUF_SIZE)
	    {
	       /* read a bufferful */
	       code = fread(buf, 1, BUF_SIZE, in);
	       if(code != BUF_SIZE)
	       {
		  fprintf(stderr, "Error: could not read from file!!!\n");
		  error++;
		  break;
	       }
	       dsize -= BUF_SIZE;
	       
	       /* write it out */
/*	       code = write(out, buf, BUF_SIZE); */
	       code = fwrite(buf, 1, BUF_SIZE, out);
	       if(code != BUF_SIZE)
	       {
		  fprintf(stderr, "Error: could not write to file!!!\n");
		  error++;
		  break;
	       }
	    }

	    if(error)
	    {
	       break;
	    }
	    
	    /* read the last block */
	    code = fread(buf, 1, dsize, in);
	    if(code != dsize)
	    {
	       fprintf(stderr, "Error: could not read from file!!!\n");
	       break;
	    }
	    /* write the last block */
/*	    code = write(out, buf, dsize); */
	    code = fwrite(buf, 1, dsize, out);
	    if(code != dsize)
	    {
	       fprintf(stderr, "Error: could not write to file!!!\n");
	       break;
	    }
	    
	 }
	 else
	 {
	    fprintf(stderr, "Error: volume %s, chunk %ld -- header corrupted!\n", vname, ch_num);
	    continue;
	 }
      }
      else /* wrong chunk, skip the data */
      {
	 if(fileseek(in, dsize, SEEK_CUR) == -1)
	 {
	    fprintf(stderr, "Error: could not seek the input file!\n");
	    break;
	 }

	 if(strcmp(fname, "-") != 0)
	 {
/*	    close(out); */
	    if(!afs)
	    {
	       if(out)
		  fclose(out);
	    }
	    else
	    {
	       if(out)
		  pclose(out);
	       afs = 0;
	    }
	 }
/*	 out = -1; */
	 out = NULL;
      }
      
      /* checks for last_vol && last_chunk */
      if(last_vol && last_chunk)
      {
	 done = 1;
      }
      /* are we done with this volume? */
      if(last_chunk)
      {

	 fprintf(stderr, "Done with %s\n", vname);
	 
	 last_chunk = 0;
	 chunk_number = 0;

	 if(strcmp(fname, "-") != 0)
	 {
/*	    close(out); */
	    if(!afs)
	    {
	       if(out)
		  fclose(out);
	    }
	    else
	    {
	       if(out)
		  pclose(out);
	       afs = 0;
	    }
	 }
/*	 out = -1; */
	 out = NULL;
      }

      /* free memory, to make sure we're not leaking */
      if(header)
      {
	 free(header);
      }
      if(vname)
      {
	 free(vname);
      }
   }

   if(buf) free(buf);
/*   if(error && out != -1) */
   if(error && out != NULL)
   {
/*      close(out); */
      if(!afs)
      {
	 fclose(out);
      }
      else
      {
	 pclose(out);
	 afs = 0;
      }
   }
   if(!done) return -1;
   else
   {
      /* done with the volset, now flush the garbage at the end of in */
      buf = (ONE_BYTE_INT *) malloc(BUF_SIZE*sizeof(ONE_BYTE_INT));
      while(!feof(in))
      {
	 fread(buf, 1, BUF_SIZE, in);
      }
      free(buf);
   }
   return 0;
}

int fsearch(FILE *in, ONE_BYTE_INT *data, int size)
// Naive search for "text" throught the in file and then returns :
//   0 if not was found 
//   1 if it was found and in ptr (if !=NULL) the position where text begins.
{
   ONE_BYTE_INT *buf;
   ONE_BYTE_INT t1;
   int i;
   int code;

   buf = (ONE_BYTE_INT *) malloc(size*sizeof(ONE_BYTE_INT));
   code = fread(buf, 1, size, in);
   
   /* first try */
   if((memcmp((void *)data, (void *)buf, size)) == 0)
   {
      free(buf);
      return 1;
   }
   /* Naive search */
   while ((code = fread(&t1, 1, 1, in)) != 0) /* while not EOF */
   {
      if(code == -1) /* error */
      {
	 fprintf(stderr, "Error: Aiiiyyooo (I/O) problems!\n");
	 free(buf);
	 return -1;
      }

      for(i = 0; i<size; i++)
      {
	 buf[i] = buf[i+1];
      }
      memcpy(buf+size-1, &t1, 1);

      if((memcmp((void *)data, (void *)buf, size)) == 0)
      {
	 free(buf);
	 return 1;
      }
   }
   free(buf);
   return 0;
}

