#ifndef _VOLSET_H
#define _VOLSET_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <rx/xdr.h>
#include <rx/rx.h>
#include <lock.h>
#include <lwp.h>
#include <ubik.h>
#include <afs/param.h>
#include <afs/afsint.h>
#include <afs/nfs.h>
#include <afs/errors.h>
#include <afs/ihandle.h>
#include <afs/dir.h>
#include <afs/vnode.h>
#include <afs/volume.h>
#include <afs/partition.h>
#include <afs/fssync.h>
#include <afs/acl.h>
#include <afs/vldbint.h>
#include <afs/dirpath.h>
#include <afs/cellconfig.h>
#include <afs/volser.h>
#include <afs/volint.h>
#include <afs/assert.h> 
#include "config.h"

/* Here's a description of the volset stream format

[Vol 0, Chunk 0 Header][Vol 0 Chunk 0 Data ][V0 C1H][V0 C1D][V1 C0H][V1 C0D]
  
The [Chunk Data] fields are CHUNK_SIZE bytes long each (except for the
last block of data for the current volume, which is not padded to the
CHUNK_SIZE), and consist of the 'vos dump' output neately chopped up
into easy to handle packages.

The [Chunk Header] fields inside look as follows:

[MAGIC][VMAJ][VMIN][HS][LV][LC][CN][DS][SERV][PART][VNS][VN][HC]

MAGIC -- the magic for the chunk, defined as VOLSET_MAGIC
VMAJ  -- major version of the volset format, defined as VERSION_MAJOR 
VMIN  -- minor version of the volset format, defined as VERSION_MINOR 
HS    -- header size, the length in bytes of the entire header,
         starting with the first byte of the magic, ending with the
	 last byte of header checksum
LV    -- flag, denotes whether the volume to which this chunk belongs
         is the last in the volset stream [1 or 0]
LC    -- flag, denotes whether the current chunk is the last chunk of
         the volume currently processed   [1 or 0]
CN    -- Chunk number, or sequence number of the current chunk in the
         volume currently being dealt with. [start with 0]
DS    -- Data size, the length in bytes of the following
         [Chunk Data] field.
SERV  -- The server the volume came from  [ sin_addr.s_addr ]
PART  -- The partition the volume came from [ same as GetPartitionID ]
VNS   -- Volume name size, a field that denotes the length (in bytes) of the
         upcoming volume name field.
VN    -- Volume name, a null terminated string that is in fact the
         name of the volume currently processed.
HC    -- Header checksum. 	 
 */


/* To make sure that these things are more or less unique */

#define CHUNK_MAGIC 0xfedcba98
#define VOLSET_MAGIC 0x76543210
#define CHUNK_SIZE 32768
#define VERSION_MAJOR 0
#define VERSION_MINOR 1

/*
Major porting issue -- who knows, maybe in 5 years (sizeof(char)!=1)
*/
/*
  If experiencing a strong desire to port, this would be a good place
  to start #ifdef'ing
*/

/* Sizes of the header fields in bytes */
#define MAGIC_SIZE   4
#define VMAJ_SIZE    1
#define VMIN_SIZE    1
#define HS_SIZE      2
#define LV_SIZE      1
#define LC_SIZE      1
#define CN_SIZE      8
#define DS_SIZE      4
#define SERV_SIZE    4
#define PART_SIZE    4
#define VNS_SIZE     2
#define HC_SIZE      1
#define CHUNK_HEADER_SIZE (MAGIC_SIZE+VMAJ_SIZE+VMIN_SIZE+\
                           HS_SIZE+LV_SIZE+LC_SIZE+CN_SIZE+DS_SIZE+\
                           SERV_SIZE+PART_SIZE+VNS_SIZE+HC_SIZE)

/* Offsets of the header fields in bytes */
#define MAGIC_OFFSET 0
#define VMAJ_OFFSET  4
#define VMIN_OFFSET  5
#define HS_OFFSET    6
#define LV_OFFSET    8
#define LC_OFFSET    9
#define CN_OFFSET    10
#define DS_OFFSET    18
#define SERV_OFFSET  22
#define PART_OFFSET  26
#define VNS_OFFSET   30
#define VN_OFFSET    32

typedef struct Node_val Node;

struct vol {
      char *volname;
      afs_int32 server;
      afs_int32 partition;
      afs_uint32 volumeId[MAXTYPES];
};

struct Node_val {
      struct vol *data;
      Node *next;
};

extern int VL_ListAttributesN2();
extern int EvalVolumeSet(struct sockaddr_in server, afs_int32 partition, char *volname, Node **vols,
		  struct ubik_client *uclient);



#endif /* _VOLSET_H */

